/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Flame
Version:   1.0
Status:    Stable
Date:      August 2002
ID:        ID_FLAME
Category:  Effect
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2002.  All rights reserved.
Short:     Creates a flame effect.
Keywords:  flame, blur, fire, burn

###DESCRIPTION###
<p>The Flame class draws an animated effect that looks like a fire burning
within a graphical area.  The effect uses a large amount of CPU time, and
this increases according to the size of the area that the graphics are being
drawn to.  The flame looks best when applied to 24 bit colour displays, as
16 bit colours or less produces high levels of colour banding.</p>

<p>Once initialised, the effect will run automatically until the object is
freed from the system.</p>

###END###

*****************************************************************************/

#include <pandora/system/class.h>
#include <pandora/graphics/bitmap.h>
#include <pandora/graphics/render.h>
#include <pandora/system/module.h>
#include <math.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_FLAME 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_FLAME, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2003.  All rights reserved.",
   "June 2003",
   "Flame"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR FlameClass = 0;
static struct RGB glFirePalette[256];

/*****************************************************************************
** Flame class definition.
*/

struct Flame {
   OBJECT_HEADER
   LONG CanvasWidth;
   LONG CanvasHeight;
   LONG Flags;

   /*** Storage area ***/
   OBJECTID DrawableID;
   OBJECTID TimerID;
   UBYTE *Data;
   LONG Time;
};

#define FLF_NODRAW 0x00000001

static struct FieldDef FlameFlags[] = {
   { "NODRAW", FLF_NODRAW },
   { NULL, NULL }
};

static struct FieldArray FlameFields[] = {
   { "CanvasWidth",  0, FDF_LONG|FDF_RI, 0, NULL, NULL },
   { "CanvasHeight", 0, FDF_LONG|FDF_RI, 0, NULL, NULL },
   { "Flags",     0, FDF_LONGFLAGS|FDF_RW, (LONG)&FlameFlags, NULL, NULL },
   END_FIELD
};

static ERROR FLAME_ActionNotify(struct Flame *, struct acActionNotify *);
static ERROR FLAME_Free(struct Flame *, APTR);
static ERROR FLAME_Init(struct Flame *, APTR);
static ERROR FLAME_NewObject(struct Flame *, APTR);
static ERROR FLAME_Timer(struct Flame *, struct acTimer *);

static struct ActionArray FlameActions[] = {
   { AC_ActionNotify, FLAME_ActionNotify },
   { AC_Free,         FLAME_Free },
   { AC_Init,         FLAME_Init },
   { AC_NewObject,    FLAME_NewObject },
   { AC_Timer,        FLAME_Timer },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   LONG i;
   FLOAT j;

   KernelBase = argKernelBase;

   for (i=0; i < 256; i++) {
      glFirePalette[i].Red = i;
   }

   j = 255.0;
   for (i=255; j > 0; i--) {
      glFirePalette[i].Green = (LONG)j;
      j -= 1.7;
   }

   j = 255.0;
   for (i=255; j > 0; i--) {
      glFirePalette[i].Blue = (LONG)j;
      j -= 2.5;
   }

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&FlameClass, NULL,
      FID_BaseClassID|TLONG, ID_FLAME,
      FID_Version|TFLOAT,    VER_FLAME,
      FID_Name|TSTRING,      "Flame",
      FID_Category|TLONG,    CCF_EFFECT,
      FID_Actions|TPTR,      FlameActions,
      FID_Fields|TPTR,       FlameFields,
      FID_Size|TLONG,        sizeof(struct Flame),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (FlameClass) { Action(AC_Free, FlameClass, NULL); FlameClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Flame: ActionNotify
*/

static ERROR FLAME_ActionNotify(struct Flame *Self, struct acActionNotify *NotifyArgs)
{
   if ((NotifyArgs->ActionID IS AC_Draw) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct Bitmap *bitmap;
      OBJECTPTR drawable;
      WORD startx, starty, x, y, addx;
      LONG width, height;
      UBYTE *data;

      if (AccessObject(Self->DrawableID, 0, &drawable) != ERR_Okay) {
         return(ObjectError(Self, ERH_Draw, ERR_ExclusiveDenied));
      }

      if (GetFields(drawable, FID_Width|TLONG,  &width,
                              FID_Height|TLONG, &height,
                              FID_Bitmap|TPTR,  &bitmap,
                              TAGEND) != ERR_Okay) {
         ReleaseObject(drawable);
         return(ObjectError(Self, ERH_Draw, ERR_GetField));
      }

      startx = 0;
      if (startx < bitmap->ClipLeft) {
         width -= bitmap->ClipLeft - startx;
         startx = bitmap->ClipLeft;
      }

      if (startx + width > bitmap->ClipRight) {
         width -= bitmap->ClipRight - startx;
      }

      starty = 0;
      if (starty < bitmap->ClipTop) {
         height -= bitmap->ClipTop - starty;
         starty = bitmap->ClipTop;
      }

      if (starty + height > bitmap->ClipBottom) {
         height -= bitmap->ClipBottom - starty;
      }

      data = Self->Data;
      for (y=starty; y < height; y++) {
         data = Self->Data + ((y * Self->CanvasHeight / height) * Self->CanvasWidth);

         addx = width - startx;
         while (addx < 0) { addx += width; data++; }

         for (x=startx; x < width; x++) {
            bitmap->DrawUCRPixel(bitmap, bitmap->XOffset + x, bitmap->YOffset + y, &glFirePalette[*data]);
            addx -= Self->CanvasWidth;
            while (addx < 0) { addx += width; data++; }
         }
      }

      ReleaseObject(drawable);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Flame: Free
*/

static ERROR FLAME_Free(struct Flame *Self, APTR Void)
{
   UnsubscribeTimer(Self->Head.UniqueID);

   if (Self->Data) {
      FreeMemory(Self->Data);
      Self->Data = NULL;
   }

   /*** Remove all subscriptions to our Drawable container ***/

   if (Self->DrawableID) {
      OBJECTPTR Drawable;
      if (AccessObject(Self->DrawableID, 3000, &Drawable) IS ERR_Okay) {
         UnsubscribeAction(Drawable, NULL,  GetUniqueID(Self));
         ReleaseObject(Drawable);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Flame: NewObject
*/

static ERROR FLAME_NewObject(struct Flame *Self, APTR Void)
{
   Self->CanvasWidth = 150;
   Self->CanvasHeight = 150;
   return(ERR_Okay);
}

/*****************************************************************************
** Flame: Init
*/

static ERROR FLAME_Init(struct Flame *Self, APTR Void)
{
   OBJECTPTR drawable;

   /*** Find the Drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = OwnerID;
   }

   /*** Allocate the internal bitmap ***/

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      if (!Self->CanvasWidth)  GetField(drawable, FID_Width, FT_LONG, &Self->CanvasWidth);
      if (!Self->CanvasHeight) GetField(drawable, FID_Height, FT_LONG, &Self->CanvasHeight);
      ReleaseObject(drawable);
   }
   else return(ObjectError(Self, ERH_Init, ERR_ExclusiveDenied));

   if (AllocMemory(Self->CanvasWidth * (Self->CanvasHeight+2), NULL, &Self->Data, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Memory));
   }

   /*** Subscribe to our container's Resize action ***/

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      SubscribeActionTags(drawable, Self, AC_Resize, AC_Redimension, AC_Draw, TAGEND);
      ReleaseObject(drawable);
   }

   /*** Initialise the Timer ***/

   return(SubscribeTimer(Self->Head.UniqueID, 60));
}

/*****************************************************************************
** Flame: Timer
*/

static ERROR FLAME_Timer(struct Flame *Self, struct acTimer *Args)
{
   LONG x, y, i, j, endline, p1, p2, p3;
   UBYTE *Data, *FireData;

   FireData = Self->Data;
   endline = Self->CanvasWidth * (Self->CanvasHeight+1);

   /*** Move lines upward ***/

   j = Self->CanvasWidth * (Self->CanvasHeight-1);
   for (i=0; i < j; i++) FireData[i] = FireData[i + Self->CanvasWidth];

   /*** I'm the fire starter, twisted fire starter ***/

   Data = FireData + endline;
   for (x=0; x < Self->CanvasWidth; x++) {
      if (RandomNumber(100) > 80) Data[x] = RandomNumber(30)+30;
      else Data[x] = 255 - 60 + RandomNumber(60);
   }

   /*** Blur fire ***/

   Data = FireData;
   for (y=Self->CanvasHeight; y > 1; y--) {
      p1 = y * Self->CanvasWidth;
      p2 = (y-1) * Self->CanvasWidth;
      p3 = (y+1) * Self->CanvasWidth;

      /*** Left side ***/
      Data[p1] = (Data[p2] + Data[p2+1] + Data[p1+1] + Data[p3] + Data[p3+1])/5;

      /*** Middle ***/
      for (x=1; x < Self->CanvasWidth-1; x++) {
         Data[p1 + x] = (Data[p1+x-1] + Data[p1+x+1] + Data[p3+x] + Data[p2+x] +
            Data[p2+x-1] + Data[p2+x+1] + Data[p3+x-1] + Data[p3+x+1])>>3;
      }

      /*** Right side ***/
      Data[p1+Self->CanvasWidth-1] = (Data[p1+Self->CanvasWidth-2] + Data[p3+Self->CanvasWidth-1] +
         Data[p2+Self->CanvasWidth-1] + Data[p2+Self->CanvasWidth-2] + Data[p3+Self->CanvasWidth-2])/5;
   }

   if (!(Self->Flags & FLF_NODRAW)) DelayMsg(AC_Draw, Self->DrawableID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   CanvasHeight
Short:  Defines the internal height of the flame graphic.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>Once it is initialised, the flame object allocates its own graphical
workspace for generating the effect.  The CanvasWidth and CanvasHeight fields have
default values, but you may want to set the dimensions of the area as it
has a direct effect on the speed and quality of the flame.</p>

<p>If the value is high, the quality will increase but the effect will
take longer to generate.  If the value is low then the quality will
decrease but the effect will be generated more quickly.  It is
recommended that the value you specify is between the range of 50 and
400.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   CanvasWidth
Short:  Defines the internal width of the flame graphic.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>Once it is initialised, the flame object allocates its own graphical
workspace for generating the effect.  The CanvasWidth and CanvasHeight fields have
default values, but you may want to set the dimensions of the area as it
has a direct effect on the speed and quality of the flame.</p>

<p>If the value is high, the quality will increase but the effect will
take longer to generate.  If the value is low then the quality will
decrease but the effect will be generated more quickly.  It is
recommended that the value you specify is between the range of 50 and
400.</p>

###END###

*****************************************************************************/
