/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      TabFocus
Version:   1.0
ID:        ID_TABFOCUS
Status:    Stable
Date:      January 2004
Category:  GUI
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2003-2004.  All rights reserved.
Keywords:  tab, focus, move, shift, active, user
Short:     Manages the 'user focus' via the tab key.

###DESCRIPTION###
<p>The TabFocus class manages the use of the tab key and its relationship to
the user-focus in the GUI of each application.  To create a tab-list, you
only need to pass a list of focus-able objects that are in the application
window.  As the user presses the tab-key, the focus will switch to each
drawable in the list, following the order that you have specified.</p>

<p>Objects can be added to the tabfocus using the <a
href="#AddTabObject()">AddTabObject</a> method for individual objects, or you can
set the Objects field for a mass addition. Some GUI classes such as the <a
href="button.html">Button</a> support a TabFocus field that you can set and
this will cause it to automatically add itself to the referenced
tabfocus object.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/input/keyboard.h>
#include <pandora/graphics/render.h>
#include <pandora/graphics/tabfocus.h>
#include <pandora/graphics/text.h>
#include <pandora/modules/render.h>
#include <pandora/functions/strtoint.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_TABFOCUS  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_TABFOCUS, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003-2004.  All rights reserved.",
   "January 2004",
   "TabFocus"
};

static struct KernelBase *KernelBase;
static struct RenderBase *RenderBase;
static OBJECTPTR TabFocusClass = NULL;
static OBJECTPTR RenderMod = NULL;

/*****************************************************************************
** TabFocus class definition.
*/

struct TabFocus {
   OBJECT_HEADER           /* Class header */
   OBJECTID DrawableID;    /* The drawable to monitor for the primary focus when managing the tab key */
   LONG     Total;         /* Total number of objects on the tab list */
   OBJECTID InputID;       /* Keyboard */
   LONG     Flags;         /* Optional flags */

   PRIVATE_FIELDS
   struct {
      OBJECTID ObjectID;
      OBJECTID DrawableID;
   } TabList[50];  /* List of objects to be managed by the tab key */
   WORD KeyboardChannel;
   WORD Index;            /* Current focus index */
};

#define TF_LOCALFOCUS 0x00000001
#define TF_CHILDFOCUS 0x00000002

static ERROR SET_Object(struct TabFocus *, STRING);
static ERROR SET_Objects(struct TabFocus *, STRING);

static struct FieldDef TabFlags[] = {
   { "LOCALFOCUS", TF_LOCALFOCUS },
   { "CHILDFOCUS", TF_CHILDFOCUS },
   { NULL, NULL }
};

static struct FieldArray TabFocusFields[] = {
   { "Drawable", 0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Total",    0, FDF_LONG|FDF_R,       0, NULL, NULL },
   { "Input",    0, FDF_OBJECTID|FDF_RI,  ID_KEYBOARD, NULL, NULL },
   { "Flags",    0, FDF_LONGFLAGS|FDF_RW, (LONG)&TabFlags, NULL, NULL },
   /*** Virtual fields ***/
   { "Objects",  0, FDF_STRING|FDF_W,     0, NULL, SET_Objects },
   { "Object",   0, FDF_STRING|FDF_W,     0, NULL, SET_Object },
   END_FIELD
};

/*****************************************************************************
** TabFocus actions
*/

static ERROR TABFOCUS_ActionNotify(struct TabFocus *, struct acActionNotify *);
static ERROR TABFOCUS_Activate(struct TabFocus *, APTR);
static ERROR TABFOCUS_Clear(struct TabFocus *, APTR);
static ERROR TABFOCUS_DataChannel(struct TabFocus *, struct acDataChannel *);
static ERROR TABFOCUS_Focus(struct TabFocus *, APTR);
static ERROR TABFOCUS_Init(struct TabFocus *, APTR);
static ERROR TABFOCUS_LostFocus(struct TabFocus *, APTR);

static struct ActionArray TabFocusActions[] = {
   { AC_ActionNotify, TABFOCUS_ActionNotify },
   { AC_Activate,     TABFOCUS_Activate },
   { AC_Clear,        TABFOCUS_Clear },
   { AC_DataChannel,  TABFOCUS_DataChannel },
   { AC_Focus,        TABFOCUS_Focus },
   { AC_Init,         TABFOCUS_Init },
   { AC_LostFocus,    TABFOCUS_LostFocus },
   { NULL, NULL }
};

/*****************************************************************************
** TabFocus methods
*/

static ERROR TABFOCUS_AddObject(struct TabFocus *, struct mtAddTabObject *);
static ERROR TABFOCUS_RemoveObject(struct TabFocus *, struct mtRemoveTabObject *);
static ERROR TABFOCUS_InsertObject(struct TabFocus *, struct mtInsertTabObject *);
static ERROR TABFOCUS_SetObject(struct TabFocus *, struct mtSetTabObject *);

static struct FunctionField argsAddObject[]    = { { "Object", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsInsertObject[] = { { "Index", ARG_LONG }, { "Object", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsRemoveObject[] = { { "Object", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsSetObject[]    = { { "Index", ARG_LONG },  { "Object", ARG_LONG }, { NULL, NULL } };

static struct MethodArray TabFocusMethods[] = {
   { MT_AddTabObject,    TABFOCUS_AddObject,    "AddObject",    argsAddObject,    sizeof(struct mtAddTabObject) },
   { MT_InsertTabObject, TABFOCUS_InsertObject, "InsertObject", argsInsertObject, sizeof(struct mtInsertTabObject) },
   { MT_RemoveTabObject, TABFOCUS_RemoveObject, "RemoveObject", argsRemoveObject, sizeof(struct mtRemoveTabObject) },
   { MT_SetTabObject,    TABFOCUS_SetObject,    "SetObject",    argsSetObject,    sizeof(struct mtSetTabObject) },
   { NULL, NULL, NULL, NULL }
};

static void FocusObject(struct TabFocus *, UBYTE Index);
static OBJECTID ObjectDrawable(OBJECTID);

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &RenderMod, NULL,
      FID_Name|TSTRING,   "render",
      FID_Version|TFLOAT, MODVERSION_RENDER,
      TAGEND) IS ERR_Okay) {
      if (GetField(RenderMod, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&TabFocusClass, NULL,
      FID_BaseClassID|TLONG, ID_TABFOCUS,
      FID_Name|TSTRING,      "TabFocus",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      TabFocusActions,
      FID_Methods|TPTR,      TabFocusMethods,
      FID_Fields|TPTR,       TabFocusFields,
      FID_Size|TLONG,        sizeof(struct TabFocus),
      FID_Version|TFLOAT,    VER_TABFOCUS,
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (RenderMod)     { Action(AC_Free, RenderMod, NULL);     RenderMod = NULL; }
   if (TabFocusClass) { Action(AC_Free, TabFocusClass, NULL); TabFocusClass = 0; }
   return(ERR_Okay);
}

/*****************************************************************************
** TabFocus: ActionNotify
*/

static ERROR TABFOCUS_ActionNotify(struct TabFocus *Self, struct acActionNotify *NotifyArgs)
{
   if (NotifyArgs->Error != ERR_Okay) return(ERR_Okay);

   if (NotifyArgs->ActionID IS AC_Focus) {
      struct RenderInfo info;
      OBJECTID currentfocus;
      BYTE focus;

      focus = FALSE;
      if (Self->Flags & (TF_LOCALFOCUS|TF_CHILDFOCUS)) {
         /* In LOCALFOCUS mode, for the tab-focus to activate, the drawable that
         ** has the user's focus must be our monitored drawable object.  In
         ** CHILDFOCUS mode, the drawable can be the monitored drawable, or
         ** it can be a child of our monitored drawable.
         */

         currentfocus = GetUserFocus();
         if (currentfocus IS Self->DrawableID) {
            Action(AC_Focus, Self, NULL);
         }
         else if (Self->Flags & TF_CHILDFOCUS) {
            if (GetRenderInfo(currentfocus, &info, sizeof(info)) IS ERR_Okay) {
               if (info.ParentID IS Self->DrawableID) {
                  focus = TRUE;
               }
            }
         }
      }
      else {
         /* In normal mode, the current focus can lie anywhere in
         ** the vicinity of the focus-path.
         */
         focus = TRUE;
      }

      if (focus) {
         OBJECTPTR input;

         if (Self->KeyboardChannel IS FALSE) {
            if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
               SubscribeChannel(input, Self);
               Self->KeyboardChannel = TRUE;
               ReleaseObject(input);
            }
            else ObjectError(Self, ERH_Focus, ERR_AccessObject);
         }
      }
   }
   else if (NotifyArgs->ActionID IS AC_LostFocus) {
      Action(AC_LostFocus, Self, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Moves the focus to the next object in the focus list.
###END###

*****************************************************************************/

static ERROR TABFOCUS_Activate(struct TabFocus *Self, APTR Args)
{
   struct RenderInfo info;
   OBJECTID currentfocus;
   WORD i, focusindex;

   DPrintF("Activate()","[TabFocus:%d]", Self->Head.UniqueID);

   if (Self->Total < 1) return(ERR_Okay);

   if ((currentfocus = GetUserFocus())) {
      if (currentfocus IS Self->DrawableID) {
         /*** Focus on the first object in the tab list ***/
         for (i=0; i < Self->Total; i++) {
            if (Self->TabList[i].ObjectID) {
               FocusObject(Self, i);
               break;
            }
         }
      }
      else {
         focusindex = -1;
         for (i=0; i < Self->Total; i++) {
            if (Self->TabList[i].DrawableID IS currentfocus) {
               /*** Go to the next object in the tag list ***/
               i++;
               while (!Self->TabList[i].ObjectID) {
                  i++;
                  if (i >= Self->Total) i = 0;
               }
               focusindex = i;
               break;
            }
         }

         /* If the current user-selected object is not in our tab-list, reset
         ** the focus to the first object in the list.
         */

         if (focusindex IS -1) {
            focusindex = Self->Index + 1;
            if ((focusindex < 0) OR (focusindex >= Self->Total)) focusindex = 0;
         }

         /* This loop ensures that the object receiving the focus
         ** is enabled.  If there are disabled objects, we skip past
         ** them to find the first active drawable.
         */

         i = Self->Total;
         while (i > 0) {
            if (Self->TabList[focusindex].DrawableID) {
               if (GetRenderInfo(Self->TabList[focusindex].DrawableID, &info, sizeof(info)) IS ERR_Okay) {
                  if (!(info.Flags & RNF_DISABLED)) {
                     FocusObject(Self, focusindex);
                     break;
                  }
               }
            }
            else {
               FocusObject(Self, focusindex);
               break;
            }

            if (++focusindex >= Self->Total) focusindex = 0;
            i--;
         }

         Self->Index = focusindex;
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    AddTabObject
Synonym: AddObject
Short:   Adds a new object to the tab list.

###DESCRIPTION###
<p>New objects can be added to the tab list by calling this method.  The
object can be of any class type, but it must support the Focus action or it
will not be able to respond when the tab list attempts to use it.</p>

<p>Once the object is added, the user will be able to focus on it by
using the tab-key.</p>

###INPUT###
OBJECTID/ObjectID: Reference to the object to be inserted.

###END###

*****************************************************************************/

static ERROR TABFOCUS_AddObject(struct TabFocus *Self, struct mtAddTabObject *Args)
{
   OBJECTID objectid, regionid;

   if ((!Args) OR (!Args->ObjectID)) return(ObjectError(Self, MT_AddTabObject, ERR_Args));

   objectid = Args->ObjectID;
   regionid = ObjectDrawable(objectid);

   /*** Do not allow references to our monitored drawable ***/

   if (objectid IS Self->DrawableID) return(ObjectError(Self, MT_AddTabObject, ERR_Failed));

   Self->TabList[Self->Total].ObjectID = objectid;
   Self->TabList[Self->Total].DrawableID = regionid;

   Self->Total++;
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Clear
Short: Flushes the tab list.
###END###

*****************************************************************************/

static ERROR TABFOCUS_Clear(struct TabFocus *Self, APTR Args)
{
   Self->Total = 0;
   return(ERR_Okay);
}

/*****************************************************************************
** TabFocus: DataChannel
*/

static ERROR TABFOCUS_DataChannel(struct TabFocus *Self, struct acDataChannel *Args)
{
   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      struct dcKeyEntry *key;
      WORD i;

      if ((key = (struct dcKeyEntry *)Args->Buffer)) {
         for (i=0; i < Args->TotalEntries; i++) {
            if (key[i].Flags & KQ_PRESSED) {
               if (key[i].Value IS K_TAB) {
                  Action(AC_Activate, Self, NULL);
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Tells the tab-focus object that the monitored drawable has received the focus.
###END###

*****************************************************************************/

static ERROR TABFOCUS_Focus(struct TabFocus *Self, APTR Void)
{
   return(ActionMsg(AC_Focus, Self->DrawableID, NULL));
}

/*****************************************************************************
** TabFocus: Init
*/

static ERROR TABFOCUS_Init(struct TabFocus *Self, APTR Void)
{
   struct Render *drawable;
   OBJECTID owner_id;

   /*** Find our parent drawable ***/

   if (!Self->DrawableID) {
      owner_id = GetOwner(Self);
      while ((owner_id) AND (GetClassID(owner_id) != ID_RENDER)) {
         owner_id = GetOwnerID(owner_id);
      }
      if (owner_id) Self->DrawableID = owner_id;
      else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
   }

   /*** Setup keyboard ***/

   if (!Self->InputID) {
      if (FastFindObject("SystemKeyboard", NULL, &Self->InputID, 1, NULL) != ERR_Okay) {
         DPrintF("@Init:","[TabFocus] Could not find any SystemKeyboard objects.");
         return(ERR_Search);
      }
   }

   if (AccessObject(Self->DrawableID, 5000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
      SubscribeActionTags(drawable, Self, AC_Focus, AC_LostFocus, TAGEND);

      if (drawable->Flags & RNF_HASFOCUS) {
         struct acActionNotify notify;
         notify.Error    = ERR_Okay;
         notify.ActionID = AC_Focus;
         TABFOCUS_ActionNotify(Self, &notify);
      }

      ReleaseObject(drawable);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    InsertTabObject
Synonym: InsertObject
Short:   Inserts a new object in the tab list.

###DESCRIPTION###
<p>New objects can be inserted into the tab list by calling this method.
You need to provide the unique ID for a drawable object, or the routine may
fail. Some intelligence is used when non-drawable objects are passed to this
method, whereby the routine will check for Region and Drawable fields to
discover valid drawable objects.</p>

<p>Once the object is inserted, the user will be able to focus on it by
using the tab-key.</p>

###INPUT###
LONG/Index: The index at which the object should be inserted.
OBJECTID/ObjectID: The ID of the object that you want to insert.

###RESULT###
ERR_Okay: The method succeeded.
ERR_Args: Invalid arguments were specified.
ERR_OutOfRange: The index is out of range.

###END###

*****************************************************************************/

static ERROR TABFOCUS_InsertObject(struct TabFocus *Self, struct mtInsertTabObject *Args)
{
   OBJECTID objectid, regionid;
   WORD index;

   if (!Args) return(ObjectError(Self, MT_InsertTabObject, ERR_Args));

   index = Args->Index;
   if (index < 0) index = 0;
   if (index >= ARRAYSIZE(Self->TabList)) return(ERR_OutOfRange);

   if ((!Args) OR (!Args->ObjectID)) return(ObjectError(Self, MT_InsertTabObject, ERR_Args));

   objectid = Args->ObjectID;
   regionid = ObjectDrawable(objectid);

   /*** Do not allow references to our monitored drawable ***/

   if (regionid IS Self->DrawableID) return(ERR_Failed);

   if (index < Self->Total) {
      CopyMemory(Self->TabList + Self->Total, Self->TabList + Self->Total + 1, sizeof(Self->TabList[0]) * (Self->Total - index));
   }

   Self->TabList[index].ObjectID = objectid;
   Self->TabList[index].DrawableID = regionid;

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  LostFocus
Short: Removes the focus from the tab and disables key monitoring.
###END###

*****************************************************************************/

static ERROR TABFOCUS_LostFocus(struct TabFocus *Self, APTR Void)
{
   OBJECTPTR input;

   if (Self->KeyboardChannel IS TRUE) {
      if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
         UnsubscribeChannel(input, Self->Head.UniqueID);
         Self->KeyboardChannel = FALSE;
         ReleaseObject(input);
      }
      else ObjectError(Self, ERH_LostFocus, ERR_AccessObject);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    RemoveTabObject
Synonym: RemoveObject
Short:   Removes an object from the tab list.

###DESCRIPTION###
<p>Use the RemoveObject() method in instances where you need to remove an existing
object from the tab list.  You only need to provide this method with the ID of
the object that you want to remove.</p>

###INPUT###
OBJECTID/ObjectID: A reference to the object that you want to remove.

###RESULT###
ERR_Okay: The object was removed, or did not already exist in the tab list.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static ERROR TABFOCUS_RemoveObject(struct TabFocus *Self, struct mtRemoveTabObject *Args)
{
   WORD i;

   if ((!Args) OR (!Args->ObjectID)) return(ObjectError(Self, MT_RemoveTabObject, ERR_Args));

   for (i=Self->Total-1; i > 0; i--) {
      if (Self->TabList[i].ObjectID IS Args->ObjectID) {
         Self->TabList[i].ObjectID = NULL;
         Self->TabList[i].DrawableID = NULL;
         CopyMemory(Self->TabList + i + 1, Self->TabList + i, (Self->Total - i - 1) * sizeof(Self->TabList[0]));
         Self->Total--;
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    SetTabObject
Synonym: SetObject
Short:   Changes the object for a specific index.

###DESCRIPTION###
<p>The tab list may be manipulated on a more direct basis by using the
SetObject() method.  It allows you to change the object ID for a specific
index point.</p>

###INPUT###
LONG/Index: The index in the tab list that you want to set.
OBJECTID/ObjectID: The ID of the object that you want to set at the specified Index.

###RESULT###
ERR_Okay: The method succeeded.
ERR_Args: Invalid arguments were specified.
ERR_OutOfRange: The specified Index was out of range.

###END###

*****************************************************************************/

static ERROR TABFOCUS_SetObject(struct TabFocus *Self, struct mtSetTabObject *Args)
{
   WORD index;

   if ((!Args) OR (!Args->ObjectID)) return(ObjectError(Self, MT_SetTabObject, ERR_Args));

   index = Args->Index;
   if (index < 0) index = 0;
   if (index >= ARRAYSIZE(Self->TabList)) return(ERR_OutOfRange);

   Self->TabList[index].ObjectID = Args->ObjectID;
   Self->TabList[index].DrawableID = ObjectDrawable(Args->ObjectID);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Object
Short:  New objects may be set at specific indexes via this field.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The Object field provides a field-based way of setting objects at specific
indexes.  An object can be set by specifying the index number, followed by
the ID of the object that you want to set.  The following is a valid example,
"3:-9495".  The index and object ID can be separated with any type of
white-space or non-numeric character(s).</p>

###END###

*****************************************************************************/

static ERROR SET_Object(struct TabFocus *Self, STRING Value)
{
   WORD index;
   OBJECTID objectid;

   while ((*Value < '0') OR (*Value > '9')) Value++; /* Find the index value */
   index = StrToInt(Value); /* Translate it */
   while ((*Value >= '0') AND (*Value <= '9')) Value++; /* Skip the index value */

   objectid = StrToInt(Value); /* Get the object ID */

   if (objectid) {
      ActionTags(MT_SetTabObject, Self, index, objectid);
      return(ERR_Okay);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###FIELD###
Name:   Objects
Short:  A string sequence of objects to be added to the tab list may be set here.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>A string sequence of objects may be added to the tab list via this field.
Objects must be specified as ID's and be separated with white-space or non-numeric
characters.  The following example illustrates a valid string "7984,
#9493, -4001".</p>

###END###

*****************************************************************************/

static ERROR SET_Objects(struct TabFocus *Self, STRING Value)
{
   struct mtAddTabObject add;
   OBJECTID objectid;

   while (*Value) {
      while ((*Value) AND (*Value != '-') AND ((*Value < '0') OR (*Value > '9'))) Value++;
      objectid = StrToInt(Value);

      if (*Value IS '-') Value++;
      while ((*Value >= '0') AND (*Value <= '9')) Value++;

      if (objectid) {
         add.ObjectID = objectid;
         TABFOCUS_AddObject(Self, &add);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: FocusObject()
*/

static void FocusObject(struct TabFocus *Self, UBYTE Index)
{
   struct mtSelectTextArea select;
   OBJECTPTR input, text;
   CLASSID class_id;

   class_id = GetClassID(Self->TabList[Index].ObjectID);

   if (class_id IS ID_INPUT) {
      if (AccessObject(Self->TabList[Index].ObjectID, 1000, &input) IS ERR_Okay) {
         Action(AC_Focus, input, NULL);

         if ((GetField(input, FID_InputText, FT_POINTER, &text) IS ERR_Okay) AND (text)) {
            select.Row = 0;
            select.Column = 0;
            select.EndRow = 200000;
            select.EndColumn = 200000;
            Action(MT_SelectTextArea, text, &select);
         }
         ReleaseObject(input);
      }
   }
   else if (ActionMsg(AC_Focus, Self->TabList[Index].ObjectID, NULL) != ERR_Okay) {
      ActionMsg(AC_Focus, Self->TabList[Index].DrawableID, NULL);
   }
/*   
   else if (class_id IS ID_COMBOBOX) ActionMsg(AC_Focus, Self->TabList[Index].ObjectID, NULL);
   else if (Self->TabList[Index].DrawableID) ActionMsg(AC_Focus, Self->TabList[Index].DrawableID, NULL);
   else ActionMsg(AC_Focus, Self->TabList[Index].ObjectID, NULL);
*/
}

/*****************************************************************************
** Internal: ObjectDrawable()
*/

static OBJECTID ObjectDrawable(OBJECTID ObjectID)
{
   OBJECTPTR object;
   OBJECTID regionid;

   if (GetClassID(ObjectID) != ID_RENDER) {
      /* If the referenced object is not a drawable, check for a region field
      ** and try to use that instead.
      */

      if (AccessObject(ObjectID, 3000, &object) IS ERR_Okay) {
         regionid = NULL;
         if (GetField(object, FID_Region, FT_LONG, &regionid) IS ERR_Okay) {
            if (GetClassID(regionid) != ID_RENDER) regionid = NULL;
         }

         if (!regionid) {
            if (GetField(object, FID_Drawable, FT_LONG, &regionid) IS ERR_Okay) {
               if (GetClassID(regionid) != ID_RENDER) regionid = NULL;
            }
         }

         ReleaseObject(object);
         return(regionid);
      }
      else return(NULL);
   }
   else return(ObjectID);
}

